"""
Set of physical constants useful to compute weather/atmosphere related formalisms
"""

air_heat_capacity = 1006
"""[J kg-1 K-1] Isobaric heat capacity (Cp) of atmosphere.

References: https://www.engineeringtoolbox.com/air-specific-heat-capacity-d_705.html
"""

air_molar_mass = 28.9654
"""[g mol-1] Average molar mass of air.

References: https://en.wikipedia.org/wiki/Density_of_air
"""

gas_constant = 8.3144598
"""[J K-1 mol-1] Universal gas constant

References: scipy.constants.gas_constant
"""

gs_max = 1 / 70.0
"""[s-1 m] Theoretical maximal value for grass surface conductance

References: box 5 in Allen et al., 1998
"""

solar_cst = 1353.
"""[W m-2] mean amount of solar radiations received by earth above the atmosphere

References: https://www.pveducation.org/pvcdrom/properties-of-sunlight/solar-radiation-outside-the-earths-atmosphere
"""

standard_atmosphere = 101325
"""[Pa] Atmospheric pressure

References: scipy.constants.physical_constants['standard atmosphere']
"""

stephan_boltzmann = 5.670367e-08
"""[W m-2 K-4] Stefan-Boltzmann constant per surface area (σ).

References: scipy.constants.physical_constants['Stefan-Boltzmann constant']
"""

water_latent_heat_vaporization = 2264.705
"""[kJ kg-1] Latent heat of vaporization of water (λ).

References: https://en.wikipedia.org/wiki/Latent_heat
"""


def air_density(t_air, atm_pressure=standard_atmosphere):
    """Volumetric mass of dry atmosphere

    References: https://en.wikipedia.org/wiki/Density_of_air

    Args:
        t_air (float): [°C] air temperature
        atm_pressure (float): [Pa]

    Returns:
        (float): [kg m-3]
    """
    t = t_air + 273.15  # convert from [°C] to [K]
    spec_cte = gas_constant / air_molar_mass * 1e3  # [J kg-1 K-1] specific gaz constant for dry air
    return atm_pressure / (spec_cte * t)



"""
Formalisms related to stomata opening
"""


def fvpd(vpd_leaf, a1, b1):
    """Effect of VPD on stomata aperture.

    References: eqS10, modified to force [0-1] range

    Args:
        vpd_leaf (float): [kPa] leaf-to-air vapour pressure difference
        a1 (float): [-] empirical factor
        b1 (float): [kPa-1] empirical factor

    Returns:
        (float): [-]
    """
    fac = a1 - b1 * vpd_leaf
    if fac <= 0:
        return 0

    if fac >= 0.5:
        return 1

    return 1 / (1 / fac - 1)


def gs_C02(an, rd, ci, ci_star, g0, fvpd):
    """stomatal conductance for CO2 diffusion.

    References: eqS3a plus forcing of g0 as min

    Args:
        an (float): [µmol CO2.m-2.s-1] net photosynthesis rate
        rd (float):[µmol CO2.m-2.s-1] mitochondrial respiration rate in the light
        ci (float): [µbar]  intercellular CO2 partial pressure
        ci_star (float): [µbar] intercellular CO2 compensation point
        g0 (float): [µmol CO2.m-2.s-1.µbar-1] residual stomatal conductance for CO2
                    at the light compensation point
        fvpd (float): [-] vapour pressure deficit effect on stomatal conductance

    Returns:
        (float): [µmol CO2.m-2.s-1.µbar-1]
    """
    if an + rd <= 0:
        return g0

    return g0 + (an + rd) / (ci - ci_star) * fvpd



"""
Actual photosynthesis formalisms
"""
import math

from . import stomata
from .constants import standard_atmosphere
from .photo_params import arrhenius, jmax
from .weather import co2_partial_pressure, vpd_leaf


def electron_transport_rate(ppfd, jmax, k2ll, convexity):
    """Rate of electron transport (J).

    References: eqS16

    Args:
        ppfd (float): [µmol photon.m-2.s-1] PAR
        jmax (float): [µmol e-.m-2.s-1] maximum electron transport rate at
                      saturating light levels
        k2ll (float): [mol e-.mol-1 photons] conversion efficiency of incident
                      PAR into electron transport at limiting light levels
        convexity (float): [-] convexity factor

    Returns:
        (float): [µmol e-.m-2.s-1] electron transport rate (J)
    """
    fac = k2ll * ppfd + jmax

    num = fac - math.sqrt(fac ** 2 - 4 * convexity * jmax * k2ll * ppfd)
    return num / (2 * convexity)


def photo_net(t_leaf, ppfd, t_atm, rh):
    """Net photosynthesis rate.

    Args:
        t_leaf (float): [°C] temperature of leaf
        ppfd (float): [?]
        t_atm (float): [°C] temperature of atmosphere
        rh (float): [-] relative humidity in the atmosphere

    Returns:
        (float): [µmol CO2.m-2.s-1]
    """
    oxygen = 210  # [mbar] oxygen partial pressure of the air
    co2 = 440  # [ppm] or [µbar.bar-1] CO2 concentration in atmosphere
    # g0 = 0.012  # [mol.m-2.s-1] TODO problem with unit [µmol CO2.m-2.s-1.µbar-1]
    g0 = 0.01  # [mol.m-2.s-1.bar-1] eq [µmol CO2.m-2.s-1.µbar-1] residual stomatal conductance for CO2 at the light compensation point from yin2009 fig6
    # gm = 0.15  # [mol.m-2.s-1] TODO problem with unit
    gm = 0.4  # [mol.m-2.s-1.bar-1] eq [µmol CO2.m-2.s-1.µbar-1] mesophyll conductance for CO2 diffusion from yin2009 fig6
    a1 = 0.84  # [-] Zhu 2018, table S1
    a1 = 0.9  # [-] yin2009 fig6
    b1 = 0.14  # [kPa-1] Zhu 2018, table S1
    b1 = 0.15  # [kPa-1] yin2009 fig6

    k2ll = 0.2  # [mol e-.mol-1 photons] Zhu 2018, table S1 with unit correction
    k2ll = 0.33 * 0.75  # [mol e-.mol-1 photons] Yin2009 table2
    convexity = 0.87  # [-] Zhu 2018, table S1
    convexity = 0.7  # [-] Zhu 2018, table S1

    jmax25 = 80 * 1.5  # [µmol e-.m-2 leaf.s-1] gs_sim
    jmax_act = 63500  # [J.mol-1] Zhu 2018, table S1
    jmax_deact = 202900  # [J.mol-1] Zhu 2018, table S1
    jmax_s = 650  # [J.K-1.mol-1] Zhu 2018, table S1

    vcmax_25 = 80  # [µmol m-2.s-1] from gs_sim  # TODO all form gs_sim can be computed from zhu by estimating N
    vcmax_act = 87700  # [J.mol-1] Zhu 2018, table S1

    kmc_25 = 27.238 * 1e6 / standard_atmosphere  # [Pa] to [µbar] Zhu 2018, table S1
    kmc_act = 80990  # [J.mol-1]

    kmo_25 = 16.582 * 1e3 / standard_atmosphere  # [Pa] to [mbar] Zhu 2018, table S1
    kmo_act = 23720  # [J.mol-1]

    rd_25 = 0.649  # [µmol m-2.s-1] from gs_sim
    rd_act = 46390  # [J.mol-1] Zhu 2018, table S1

    sco_25 = 2.8  # [mbar.µbar-1] from [kPa.Pa-1] Zhu 2018, table S1
    sco_act = -24460  # [J.mol-1]

    kmc = arrhenius(t_leaf, kmc_25, kmc_act)  # needed [µbar]
    print("kmc", kmc)
    kmo = arrhenius(t_leaf, kmo_25, kmo_act)  # needed [mbar]
    print("kmmo", kmo)
    sco = arrhenius(t_leaf, sco_25, sco_act)  # needed [mbar.µbar-1]
    print("sco", sco)

    # rtb = 2.357 / 1.6  # [mol CO2.m-2.s-1] TODO pb with unit  # [m2.s.µbar.µmol-1]  combined turbulence and boundary layer resistance for CO2
    # print("gb", 1 / rtb)
    gb = 1.5  # [mol.m-2.s-1.bar-1] from yin2009 fig6
    print("gb", gb)
    rtb = 1 / gb

    fvpd = stomata.fvpd(vpd_leaf(t_leaf, t_atm, rh), a1, b1)  # [-]
    print("vpd_leaf", vpd_leaf(t_leaf, t_atm, rh))
    vcmax = arrhenius(t_leaf, vcmax_25, vcmax_act)  # needed [µmol CO2.m-2.s-1]
    vcmax = 120  # [µmol CO2.m-2.s-1]  Yin2009 fig6
    print("vcmax", vcmax)
    rd = arrhenius(t_leaf, rd_25, rd_act)  # needed [µmol CO2.m-2.s-1]
    rd = 0.01 * vcmax  # [µmol CO2.m-2.s-1]  Yin2009 fig6
    rd = 0
    print("rd", rd)
    ca = co2_partial_pressure(t_atm, co2)  # [µbar]
    ca = 360  # [µbar]  Yin2009 fig6
    print("ca", ca)

    jmax_val = jmax(t_leaf, jmax25, jmax_act, jmax_deact, jmax_s)
    jmax_val = 230  # [µmol CO2.m-2.s-1]  Yin2009 fig6
    print("jmax", jmax_val)
    j_maj = electron_transport_rate(ppfd, jmax_val, k2ll, convexity)  # [µmol e-.m-2.s-1]

    gamma_star = 0.5 * oxygen / sco  # eqS12 needed [µbar]

    # x1 [µmol CO2.m-2.s-1] eqS15a
    # x2 [µbar] eqS15b
    ans = []
    for x1, x2 in [(vcmax, kmc * (1 + oxygen / kmo)),  # rubisco limited
                   (j_maj / 4, 2 * gamma_star)]:  # e- transport limited
        a = g0 * (x2 + gamma_star) + (g0 / gm + fvpd) * (x1 - rd)  # eqS6a
        b = ca * (x1 - rd) - gamma_star * x1 - rd * x2  # eqS6b
        c = ca + x2 + (1 / gm + rtb) * (x1 - rd)  # eqS6c with (rtb = rt + rb)
        d = x2 + gamma_star + (x1 - rd) / gm  # eqS6d
        e = 1 / gm + (g0 / gm + fvpd) * (1 / gm + rtb)  # eqS6e with (rtb = rt + rb)

        p = -(d + (x1 - rd) / gm + a * (1 / gm + rtb) + (g0 / gm + fvpd) * c) / e  # eqS5a with (rtb = rt + rb)
        q = (d * (x1 - rd) + a * c + (g0 / gm + fvpd) * b) / e  # eqS5b
        r = - a * b / e  # eqS5c

        q_maj = (p ** 2 - 3 * q) / 9  # eqS8a
        r_maj = (2 * p ** 3 - 9 * p * q + 27 * r) / 54  # eqS8c
        try:
            theta = math.acos(r_maj / math.sqrt(q_maj ** 3))  # eqS8b

            an = -2 * math.sqrt(q_maj) * math.cos(theta / 3 + 4 * math.pi / 3) - p / 3  # eqS7
            ans.append(an)
        except ValueError:
            raise  # TODO temporary??

    ci = ca - min(ans) * (1 / (g0 * 10) + 1 / gb)
    cc = ci - min(ans) / gm

    print("res", 1 / g0, 1 / gb, 1 / gm)
    print("ci", ci)
    print("cc", cc)
    return min(ans)  # from eqS15



"""
Weather, physical atmosphere, related formalisms
"""
import math

from .constants import gas_constant


def vapour_sat(temp):
    """Saturated vapour pressure of the air.

    References: eqS23

    Args:
        temp (float): [°C] temperature of air

    Returns:
        (float): [kPa]
    """
    return 0.611 * math.exp(17.4 * temp / (239 + temp))


def vpd(temp, rh):
    """Vapour pressure difference in the air.

    References: eqS22 + eqS24

    Args:
        temp (float): [°C] temperature of air
        rh (float): [-] relative humidity of the air

    Returns:
        (float): [kPa]
    """
    return (1 - rh) * vapour_sat(temp)


def vpd_leaf(t_leaf, t_atm, rh):
    """leaf-to-air vapour pressure difference.

    References: eqS22 + eqS24

    Args:
        t_leaf (float): [°C] temperature of leaf
        t_atm (float): [°C] temperature of air
        rh (float): [-] relative humidity of the air

    Returns:
        (float): [kPa]
    """
    return vapour_sat(t_leaf) - rh * vapour_sat(t_atm)


def co2_partial_pressure(temp, co2_conc):
    """CO2 partial pressure of the air.

    References: eqS21

    Args:
        temp (float): [°C] temperature of air
        co2_conc (float): [µmol.mol-1] CO2 concentration in the air

    Returns:
        (float): [µbar]
    """
    return co2_conc * gas_constant * 1e-3 * (273.15 + temp) * 10 / 22.4136



"""
Secondary parameters for photosynthesis
"""
import math

from .constants import gas_constant


def arrhenius(temp, v25, eact):
    """Typical temperature dependency formalism.

    References: eqS14

    Args:
        temp (float): [°C] object temperature
        v25 (float): [any]
        eact (float): [J.mol-1]

    Returns:
        (float): [same as v25]
    """
    t25 = 25 + 273.15
    tk = temp + 273.15

    return v25 * math.exp((1 / t25 - 1 / tk) * eact / gas_constant)


def _jmax_decr(jmax_s, jmax_deact, tk):
    return 1 + math.exp((jmax_s - jmax_deact / tk) / gas_constant)


def jmax(temp, jmax25, jmax_act, jmax_deact, jmax_s):
    """Maximum electron transport rate at saturating light levels.

    References: eqS17

    Notes: bug correction in original paper, use minus sign

    Args:
        temp (float): [°C] temperature
        jmax25 (float): [µmol e-.m-2 leaf.s-1] value of jmax at 25°C
        jmax_act (float): [J.mol-1] activation energy for jmax
        jmax_deact (float): [J.mol-1] deactivation energy
        jmax_s (float): [J.K-1.mol-1] entropy term

    Returns:
        (float): [µmol e-.m-2 leaf.s-1] Maximum electron transport rate (Jmax)
    """
    tk = temp + 273.15
    t25 = 25 + 273.15

    sca = _jmax_decr(jmax_s, jmax_deact, t25) / _jmax_decr(jmax_s, jmax_deact, tk)
    return jmax25 * math.exp((1 / t25 - 1 / tk) * jmax_act / gas_constant) * sca
